# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime
from datetime import timezone
from datetime import timedelta

from ooresults.plugins import iof_result_list
from ooresults.repo import result_type
from ooresults.repo.result_type import ResultStatus


def test_import_result_list():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <ClassResult>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
    <PersonResult>
      <Person sex="F">
        <Name>
          <Family>Merkel</Family>
          <Given>Angela</Given>
        </Name>
        <BirthDate>1972-01-01</BirthDate>
      </Person>
      <Organisation>
        <Name>OC Kanzleramt</Name>
      </Organisation>
      <Result>
        <StartTime>2020-02-09T10:00:00+01:00</StartTime>
        <FinishTime>2020-02-09T10:33:21+01:00</FinishTime>
        <Time>2001</Time>
        <Status>OK</Status>
        <SplitTime>
          <ControlCode>31</ControlCode>
          <Time>501</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>32</ControlCode>
          <Time>720</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>31</ControlCode>
          <Time>818</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>33</ControlCode>
          <Time>1136</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>31</ControlCode>
          <Time>1593</Time>
        </SplitTime>
        <ControlCard punchingSystem="SI">1234567</ControlCard>
      </Result>
    </PersonResult>
  </ClassResult>
</ResultList>
"""
    event, results = iof_result_list.parse_result_list(bytes(content, encoding="utf-8"))
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert results == [
        {
            "first_name": "Angela",
            "last_name": "Merkel",
            "class_": "Bahn A - Lang",
            "club": "OC Kanzleramt",
            "chip": "1234567",
            "gender": "F",
            "year": 1972,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                start_time=datetime.datetime(
                    2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                ),
                finish_time=datetime.datetime(
                    2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                ),
                punched_start_time=datetime.datetime(
                    2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                ),
                punched_finish_time=datetime.datetime(
                    2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                ),
                status=ResultStatus.OK,
                time=2001,
                split_times=[
                    result_type.SplitTime(control_code="31", status="OK", time=501),
                    result_type.SplitTime(control_code="32", status="OK", time=720),
                    result_type.SplitTime(control_code="31", status="OK", time=818),
                    result_type.SplitTime(control_code="33", status="OK", time=1136),
                    result_type.SplitTime(control_code="31", status="OK", time=1593),
                ],
            ),
        },
    ]


def test_import_result_list_not_competing():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <ClassResult>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
    <PersonResult>
      <Person sex="F">
        <Name>
          <Family>Merkel</Family>
          <Given>Angela</Given>
        </Name>
        <BirthDate>1972-01-01</BirthDate>
      </Person>
      <Organisation>
        <Name>OC Kanzleramt</Name>
      </Organisation>
      <Result>
        <StartTime>2020-02-09T10:00:00+01:00</StartTime>
        <FinishTime>2020-02-09T10:33:21+01:00</FinishTime>
        <Time>2001</Time>
        <Status>NotCompeting</Status>
        <ControlCard punchingSystem="SI">1234567</ControlCard>
      </Result>
    </PersonResult>
  </ClassResult>
</ResultList>
"""
    event, results = iof_result_list.parse_result_list(bytes(content, encoding="utf-8"))
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert results == [
        {
            "first_name": "Angela",
            "last_name": "Merkel",
            "class_": "Bahn A - Lang",
            "club": "OC Kanzleramt",
            "chip": "1234567",
            "gender": "F",
            "year": 1972,
            "not_competing": True,
            "result": result_type.PersonRaceResult(
                start_time=datetime.datetime(
                    2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                ),
                finish_time=datetime.datetime(
                    2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                ),
                punched_start_time=datetime.datetime(
                    2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                ),
                punched_finish_time=datetime.datetime(
                    2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                ),
                status=ResultStatus.OK,
                time=2001,
            ),
        },
    ]


def test_export_result_list():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <ClassResult>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
    <PersonResult>
      <Person sex="F">
        <Name>
          <Family>Merkel</Family>
          <Given>Angela</Given>
        </Name>
        <BirthDate>1972-01-01</BirthDate>
      </Person>
      <Organisation>
        <Name>OC Kanzleramt</Name>
      </Organisation>
      <Result>
        <StartTime>2020-02-09T10:00:00+01:00</StartTime>
        <FinishTime>2020-02-09T10:33:21+01:00</FinishTime>
        <Time>2001</Time>
        <Status>OK</Status>
        <SplitTime>
          <ControlCode>31</ControlCode>
          <Time>501</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>32</ControlCode>
          <Time>720</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>31</ControlCode>
          <Time>818</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>33</ControlCode>
          <Time>1136</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>31</ControlCode>
          <Time>1593</Time>
        </SplitTime>
        <ControlCard punchingSystem="SI">1234567</ControlCard>
      </Result>
    </PersonResult>
  </ClassResult>
</ResultList>
"""
    document = iof_result_list.create_result_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "club": "OC Kanzleramt",
                "chip": "1234567",
                "gender": "F",
                "year": 1972,
                "not_competing": False,
                "result": result_type.PersonRaceResult(
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                    ),
                    finish_time=datetime.datetime(
                        2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                    ),
                    status=ResultStatus.OK,
                    time=2001,
                    split_times=[
                        result_type.SplitTime(control_code="31", status="OK", time=501),
                        result_type.SplitTime(control_code="32", status="OK", time=720),
                        result_type.SplitTime(control_code="31", status="OK", time=818),
                        result_type.SplitTime(
                            control_code="33", status="OK", time=1136
                        ),
                        result_type.SplitTime(
                            control_code="31", status="OK", time=1593
                        ),
                    ],
                ),
            },
        ],
    )
    assert document == bytes(content, encoding="utf-8")


def test_export_result_list_not_competing():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <ClassResult>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
    <PersonResult>
      <Person sex="F">
        <Name>
          <Family>Merkel</Family>
          <Given>Angela</Given>
        </Name>
        <BirthDate>1972-01-01</BirthDate>
      </Person>
      <Organisation>
        <Name>OC Kanzleramt</Name>
      </Organisation>
      <Result>
        <StartTime>2020-02-09T10:00:00+01:00</StartTime>
        <FinishTime>2020-02-09T10:33:21+01:00</FinishTime>
        <Time>2001</Time>
        <Status>NotCompeting</Status>
        <ControlCard punchingSystem="SI">1234567</ControlCard>
      </Result>
    </PersonResult>
  </ClassResult>
</ResultList>
"""
    document = iof_result_list.create_result_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "club": "OC Kanzleramt",
                "chip": "1234567",
                "gender": "F",
                "year": 1972,
                "not_competing": True,
                "result": result_type.PersonRaceResult(
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                    ),
                    finish_time=datetime.datetime(
                        2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                    ),
                    status=ResultStatus.OK,
                    time=2001,
                ),
            },
        ],
    )
    assert document == bytes(content, encoding="utf-8")


def test_export_result_list_not_competing():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <ClassResult>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
    <PersonResult>
      <Person sex="F">
        <Name>
          <Family>Merkel</Family>
          <Given>Angela</Given>
        </Name>
        <BirthDate>1972-01-01</BirthDate>
      </Person>
      <Organisation>
        <Name>OC Kanzleramt</Name>
      </Organisation>
      <Result>
        <StartTime>2020-02-09T10:00:00+01:00</StartTime>
        <FinishTime>2020-02-09T10:33:21+01:00</FinishTime>
        <Time>2001</Time>
        <Position>1</Position>
        <Status>OK</Status>
        <ControlCard punchingSystem="SI">1234567</ControlCard>
      </Result>
    </PersonResult>
  </ClassResult>
</ResultList>
"""
    document = iof_result_list.create_result_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "club": "OC Kanzleramt",
                "chip": "1234567",
                "gender": "F",
                "year": 1972,
                "not_competing": False,
                "rank": 1,
                "result": result_type.PersonRaceResult(
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                    ),
                    finish_time=datetime.datetime(
                        2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                    ),
                    status=ResultStatus.OK,
                    time=2001,
                ),
            },
        ],
    )
    assert document == bytes(content, encoding="utf-8")


def test_import_result_list_without_class_result():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
</ResultList>
"""
    event, results = iof_result_list.parse_result_list(bytes(content, encoding="utf-8"))
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert results == []


def test_export_result_list_without_class_result():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
</ResultList>
"""
    document = iof_result_list.create_result_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [],
    )
    assert document == bytes(content, encoding="utf-8")


def test_import_result_list_classes():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <ClassResult>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
    <PersonResult>
      <Person>
        <Name>
          <Family>Merkel</Family>
          <Given>Angela</Given>
        </Name>
      </Person>
      <Result>
        <StartTime>2020-02-09T10:05:00Z</StartTime>
        <FinishTime>2020-02-09T10:25:21Z</FinishTime>
        <Time>1221</Time>
        <Status>OK</Status>
        <SplitTime>
          <ControlCode>41</ControlCode>
          <Time>301</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>42</ControlCode>
          <Time>526</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>41</ControlCode>
          <Time>914</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>43</ControlCode>
          <Time>1100</Time>
        </SplitTime>
      </Result>
    </PersonResult>
    <PersonResult>
      <Person>
        <Name>
          <Family>Merkel</Family>
          <Given>Birgit</Given>
        </Name>
      </Person>
      <Result>
        <StartTime>2020-02-09T10:06:00+00:00</StartTime>
        <Status>DidNotFinish</Status>
        <SplitTime>
          <ControlCode>41</ControlCode>
          <Time>501</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>42</ControlCode>
          <Time>720</Time>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>41</ControlCode>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>43</ControlCode>
        </SplitTime>
      </Result>
    </PersonResult>
  </ClassResult>
  <ClassResult>
    <Class>
      <Name>Bahn B - Mittel</Name>
    </Class>
    <PersonResult>
      <Person>
        <Name>
          <Family>Merkel</Family>
          <Given>Claudia</Given>
        </Name>
      </Person>
      <Result>
        <StartTime>2020-02-09T10:00:00Z</StartTime>
        <FinishTime>2020-02-09T10:33:21Z</FinishTime>
        <Time>2001</Time>
        <Status>MissingPunch</Status>
        <SplitTime status="OK">
          <ControlCode>31</ControlCode>
          <Time>501</Time>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>32</ControlCode>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>31</ControlCode>
        </SplitTime>
        <SplitTime status="Additional">
          <ControlCode>33</ControlCode>
          <Time>1136</Time>
        </SplitTime>
        <ControlCard punchingSystem="SI">1234567</ControlCard>
      </Result>
    </PersonResult>
  </ClassResult>
</ResultList>
"""
    event, results = iof_result_list.parse_result_list(bytes(content, encoding="utf-8"))
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert results == [
        {
            "first_name": "Angela",
            "last_name": "Merkel",
            "class_": "Bahn A - Lang",
            "club": "",
            "chip": "",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.OK,
                start_time=datetime.datetime(2020, 2, 9, 10, 5, 0, tzinfo=timezone.utc),
                finish_time=datetime.datetime(
                    2020, 2, 9, 10, 25, 21, tzinfo=timezone.utc
                ),
                punched_start_time=datetime.datetime(
                    2020, 2, 9, 10, 5, 0, tzinfo=timezone.utc
                ),
                punched_finish_time=datetime.datetime(
                    2020, 2, 9, 10, 25, 21, tzinfo=timezone.utc
                ),
                time=1221,
                split_times=[
                    result_type.SplitTime(control_code="41", status="OK", time=301),
                    result_type.SplitTime(control_code="42", status="OK", time=526),
                    result_type.SplitTime(control_code="41", status="OK", time=914),
                    result_type.SplitTime(control_code="43", status="OK", time=1100),
                ],
            ),
        },
        {
            "first_name": "Birgit",
            "last_name": "Merkel",
            "class_": "Bahn A - Lang",
            "club": "",
            "chip": "",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DID_NOT_FINISH,
                start_time=datetime.datetime(2020, 2, 9, 10, 6, 0, tzinfo=timezone.utc),
                punched_start_time=datetime.datetime(
                    2020, 2, 9, 10, 6, 0, tzinfo=timezone.utc
                ),
                split_times=[
                    result_type.SplitTime(control_code="41", status="OK", time=501),
                    result_type.SplitTime(control_code="42", status="OK", time=720),
                    result_type.SplitTime(control_code="41", status="Missing"),
                    result_type.SplitTime(control_code="43", status="Missing"),
                ],
            ),
        },
        {
            "first_name": "Claudia",
            "last_name": "Merkel",
            "class_": "Bahn B - Mittel",
            "club": "",
            "chip": "1234567",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.MISSING_PUNCH,
                start_time=datetime.datetime(2020, 2, 9, 10, 0, 0, tzinfo=timezone.utc),
                finish_time=datetime.datetime(
                    2020, 2, 9, 10, 33, 21, tzinfo=timezone.utc
                ),
                punched_start_time=datetime.datetime(
                    2020, 2, 9, 10, 0, 0, tzinfo=timezone.utc
                ),
                punched_finish_time=datetime.datetime(
                    2020, 2, 9, 10, 33, 21, tzinfo=timezone.utc
                ),
                time=2001,
                split_times=[
                    result_type.SplitTime(control_code="31", status="OK", time=501),
                    result_type.SplitTime(control_code="32", status="Missing"),
                    result_type.SplitTime(control_code="31", status="Missing"),
                    result_type.SplitTime(
                        control_code="33", status="Additional", time=1136
                    ),
                ],
            ),
        },
    ]


def test_export_result_list_classes():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ResultList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <ClassResult>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
    <PersonResult>
      <Person>
        <Name>
          <Family>Merkel</Family>
          <Given>Angela</Given>
        </Name>
      </Person>
      <Result>
        <StartTime>2020-02-09T10:05:00+00:00</StartTime>
        <FinishTime>2020-02-09T10:25:21+00:00</FinishTime>
        <Time>1221</Time>
        <Status>OK</Status>
        <SplitTime>
          <ControlCode>41</ControlCode>
          <Time>301</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>42</ControlCode>
          <Time>526</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>41</ControlCode>
          <Time>914</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>43</ControlCode>
          <Time>1100</Time>
        </SplitTime>
      </Result>
    </PersonResult>
    <PersonResult>
      <Person>
        <Name>
          <Family>Merkel</Family>
          <Given>Birgit</Given>
        </Name>
      </Person>
      <Result>
        <StartTime>2020-02-09T10:06:00+00:00</StartTime>
        <Status>DidNotFinish</Status>
        <SplitTime>
          <ControlCode>41</ControlCode>
          <Time>501</Time>
        </SplitTime>
        <SplitTime>
          <ControlCode>42</ControlCode>
          <Time>720</Time>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>41</ControlCode>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>43</ControlCode>
        </SplitTime>
      </Result>
    </PersonResult>
  </ClassResult>
  <ClassResult>
    <Class>
      <Name>Bahn B - Mittel</Name>
    </Class>
    <PersonResult>
      <Person>
        <Name>
          <Family>Merkel</Family>
          <Given>Claudia</Given>
        </Name>
      </Person>
      <Result>
        <StartTime>2020-02-09T10:00:00+00:00</StartTime>
        <FinishTime>2020-02-09T10:33:21+00:00</FinishTime>
        <Time>2001</Time>
        <Status>MissingPunch</Status>
        <SplitTime>
          <ControlCode>31</ControlCode>
          <Time>501</Time>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>32</ControlCode>
        </SplitTime>
        <SplitTime status="Missing">
          <ControlCode>31</ControlCode>
        </SplitTime>
        <SplitTime status="Additional">
          <ControlCode>33</ControlCode>
          <Time>1136</Time>
        </SplitTime>
        <ControlCard punchingSystem="SI">1234567</ControlCard>
      </Result>
    </PersonResult>
  </ClassResult>
</ResultList>
"""
    document = iof_result_list.create_result_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "not_competing": False,
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK,
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 5, 0, tzinfo=timezone.utc
                    ),
                    finish_time=datetime.datetime(
                        2020, 2, 9, 10, 25, 21, tzinfo=timezone.utc
                    ),
                    time=1221,
                    split_times=[
                        result_type.SplitTime(control_code="41", status="OK", time=301),
                        result_type.SplitTime(control_code="42", status="OK", time=526),
                        result_type.SplitTime(control_code="41", status="OK", time=914),
                        result_type.SplitTime(
                            control_code="43", status="OK", time=1100
                        ),
                    ],
                ),
            },
            {
                "first_name": "Claudia",
                "last_name": "Merkel",
                "class_": "Bahn B - Mittel",
                "chip": "1234567",
                "not_competing": False,
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.MISSING_PUNCH,
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 0, 0, tzinfo=timezone.utc
                    ),
                    finish_time=datetime.datetime(
                        2020, 2, 9, 10, 33, 21, tzinfo=timezone.utc
                    ),
                    time=2001,
                    split_times=[
                        result_type.SplitTime(control_code="31", status="OK", time=501),
                        result_type.SplitTime(control_code="32", status="Missing"),
                        result_type.SplitTime(control_code="31", status="Missing"),
                        result_type.SplitTime(
                            control_code="33", status="Additional", time=1136
                        ),
                    ],
                ),
            },
            {
                "first_name": "Birgit",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "not_competing": False,
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.DID_NOT_FINISH,
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 6, 0, tzinfo=timezone.utc
                    ),
                    split_times=[
                        result_type.SplitTime(control_code="41", status="OK", time=501),
                        result_type.SplitTime(control_code="42", status="OK", time=720),
                        result_type.SplitTime(control_code="41", status="Missing"),
                        result_type.SplitTime(control_code="43", status="Missing"),
                    ],
                ),
            },
        ],
    )
    assert document == bytes(content, encoding="utf-8")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

from tests.entry import Entry
from ooresults.plugins import oe2003
from ooresults.repo.result_type import PersonRaceResult
from ooresults.repo.result_type import ResultStatus


header = "Stnr;Chip;Datenbank Id;Nachname;Vorname;Jg;G;Block;AK;Start;Ziel;Zeit;Wertung;Club-Nr.;Abk;Ort;Nat;Katnr;Kurz;Lang"
encoding = "windows-1252"


def test_separator_semicolon():
    content = oe2003.create(
        entries=[Entry(first_name="v", last_name="n", chip="c")], class_list=[]
    )
    v1 = "1;c;;n;v;;;;0;;;;;;;;;;;"

    assert content == bytes(header + "\r\n" + v1 + "\r\n", encoding=encoding)


def test_quote_within_quotes():
    content = oe2003.create(
        entries=[Entry(first_name="v", last_name='n1"2', chip="c")], class_list=[]
    )
    v1 = '1;c;;"n1""2";v;;;;0;;;;;;;;;;;'

    assert content == bytes(header + "\r\n" + v1 + "\r\n", encoding=encoding)


def test_separator_within_quotes():
    content = oe2003.create(
        entries=[Entry(first_name="v", last_name="n1;2", chip="c")], class_list=[]
    )
    v1 = '1;c;;"n1;2";v;;;;0;;;;;;;;;;;'

    assert content == bytes(header + "\r\n" + v1 + "\r\n", encoding=encoding)


def test_newline_within_quotes():
    content = oe2003.create(
        entries=[Entry(first_name="v", last_name="n1\n2", chip="c")], class_list=[]
    )
    v1 = '1;c;;"n1\n2";v;;;;0;;;;;;;;;;;'

    assert content == bytes(header + "\r\n" + v1 + "\r\n", encoding=encoding)


def test_special_characters():
    content = oe2003.create(
        entries=[Entry(first_name="v", last_name="ÄÖÜßäüö", chip="c")], class_list=[]
    )
    v1 = "1;c;;ÄÖÜßäüö;v;;;;0;;;;;;;;;;;"

    assert content == bytes(header + "\r\n" + v1 + "\r\n", encoding=encoding)


def test_multiline():
    content = oe2003.create(
        entries=[
            Entry(first_name="v1", last_name="n1", chip="c1"),
            Entry(first_name="v2", last_name="n2", chip="c2"),
            Entry(first_name="v3", last_name="n3", chip="c3"),
        ],
        class_list=[],
    )
    v1 = "1;c1;;n1;v1;;;;0;;;;;;;;;;;"
    v2 = "2;c2;;n2;v2;;;;0;;;;;;;;;;;"
    v3 = "3;c3;;n3;v3;;;;0;;;;;;;;;;;"

    assert content == bytes(
        header + "\r\n" + v1 + "\r\n" + v2 + "\r\n" + v3 + "\r\n", encoding=encoding
    )


def test_year():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", year=2006),
            Entry(first_name="c", last_name="d", year=2000),
            Entry(first_name="e", last_name="f", year=1941),
        ],
        class_list=[],
    )
    v1 = (
        "1;;;b;a;2006;;;0;;;;;;;;;;;"
        + "\r\n"
        + "2;;;d;c;2000;;;0;;;;;;;;;;;"
        + "\r\n"
        + "3;;;f;e;1941;;;0;;;;;;;;;;;"
        + "\r\n"
    )

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_gender():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", gender="F"),
            Entry(first_name="c", last_name="d", gender="M"),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;W;;0;;;;;;;;;;;" + "\r\n" + "2;;;d;c;;M;;0;;;;;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_not_competing():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", not_competing=False),
            Entry(first_name="c", last_name="d", not_competing=True),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;;;;;;;;" + "\r\n" + "2;;;d;c;;;;X;;;;;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_club():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", club_id=1, club="OL1"),
            Entry(first_name="c", last_name="d", club_id=2, club="OL2"),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;;1;;OL1;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;;2;;OL2;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_club_not_exported_without_club_id():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", club_id=None, club="OL1"),
            Entry(first_name="c", last_name="d", club_id=2, club="OL2"),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;;;;;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;;2;;OL2;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_class():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", class_id=1, class_="Class_1"),
            Entry(first_name="c", last_name="d", class_id=2, class_="Class_2"),
        ],
        class_list=[],
    )
    v1 = (
        "1;;;b;a;;;;0;;;;;;;;;;Class_1;Class_1"
        + "\r\n"
        + "2;;;d;c;;;;0;;;;;;;;;;Class_2;Class_2"
        + "\r\n"
    )

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_class_short_name():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", class_id=1, class_="Class_1"),
            Entry(first_name="c", last_name="d", class_id=2, class_="Class_2"),
        ],
        class_list=[
            {"id": 1, "name": "Class_1", "short_name": "C1"},
            {"id": 2, "name": "Class_2", "short_name": "C2"},
        ],
    )
    v1 = (
        "1;;;b;a;;;;0;;;;;;;;;1;C1;Class_1"
        + "\r\n"
        + "2;;;d;c;;;;0;;;;;;;;;2;C2;Class_2"
        + "\r\n"
    )

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_start_time():
    s1 = datetime.datetime(
        2020, 2, 9, 10, 0, 15, tzinfo=datetime.timezone(datetime.timedelta(hours=1))
    )
    s2 = datetime.datetime(
        2020, 3, 1, 22, 50, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=1))
    )
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a", last_name="b", result=PersonRaceResult(start_time=s1)
            ),
            Entry(
                first_name="c", last_name="d", result=PersonRaceResult(start_time=s2)
            ),
        ],
        class_list=[],
    )
    v1 = (
        "1;;;b;a;;;;0;10:00:15;;;;;;;;;;"
        + "\r\n"
        + "2;;;d;c;;;;0;22:50:00;;;;;;;;;;"
        + "\r\n"
    )

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_finish_time():
    f1 = datetime.datetime(
        2020, 2, 9, 10, 0, 15, tzinfo=datetime.timezone(datetime.timedelta(hours=1))
    )
    f2 = datetime.datetime(
        2020, 3, 1, 22, 50, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=1))
    )
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a", last_name="b", result=PersonRaceResult(finish_time=f1)
            ),
            Entry(
                first_name="c", last_name="d", result=PersonRaceResult(finish_time=f2)
            ),
        ],
        class_list=[],
    )
    v1 = (
        "1;;;b;a;;;;0;;10:00:15;;;;;;;;;"
        + "\r\n"
        + "2;;;d;c;;;;0;;22:50:00;;;;;;;;;"
        + "\r\n"
    )

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_time():
    content = oe2003.create(
        entries=[
            Entry(first_name="a", last_name="b", result=PersonRaceResult(time=301)),
            Entry(first_name="c", last_name="d", result=PersonRaceResult(time=8000)),
        ],
        class_list=[],
    )
    v1 = (
        "1;;;b;a;;;;0;;;5:01;;;;;;;;"
        + "\r\n"
        + "2;;;d;c;;;;0;;;133:20;;;;;;;;"
        + "\r\n"
    )

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_status_ok():
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a",
                last_name="b",
                result=PersonRaceResult(status=ResultStatus.OK),
            ),
            Entry(
                first_name="c",
                last_name="d",
                result=PersonRaceResult(status=ResultStatus.OK),
            ),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;0;;;;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;0;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_status_dns():
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a",
                last_name="b",
                result=PersonRaceResult(status=ResultStatus.DID_NOT_START),
            ),
            Entry(
                first_name="c",
                last_name="d",
                result=PersonRaceResult(status=ResultStatus.DID_NOT_START),
            ),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;1;;;;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;1;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_status_dnf():
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a",
                last_name="b",
                result=PersonRaceResult(status=ResultStatus.DID_NOT_FINISH),
            ),
            Entry(
                first_name="c",
                last_name="d",
                result=PersonRaceResult(status=ResultStatus.DID_NOT_FINISH),
            ),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;2;;;;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;2;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_status_mp():
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a",
                last_name="b",
                result=PersonRaceResult(status=ResultStatus.MISSING_PUNCH),
            ),
            Entry(
                first_name="c",
                last_name="d",
                result=PersonRaceResult(status=ResultStatus.MISSING_PUNCH),
            ),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;3;;;;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;3;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_status_disq():
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a",
                last_name="b",
                result=PersonRaceResult(status=ResultStatus.DISQUALIFIED),
            ),
            Entry(
                first_name="c",
                last_name="d",
                result=PersonRaceResult(status=ResultStatus.DISQUALIFIED),
            ),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;4;;;;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;4;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_status_otl():
    content = oe2003.create(
        entries=[
            Entry(
                first_name="a",
                last_name="b",
                result=PersonRaceResult(status=ResultStatus.OVER_TIME),
            ),
            Entry(
                first_name="c",
                last_name="d",
                result=PersonRaceResult(status=ResultStatus.OVER_TIME),
            ),
        ],
        class_list=[],
    )
    v1 = "1;;;b;a;;;;0;;;;5;;;;;;;" + "\r\n" + "2;;;d;c;;;;0;;;;5;;;;;;;" + "\r\n"

    assert content == bytes(header + "\r\n" + v1, encoding=encoding)


def test_status_diacritic_characters_cp1252_encoding():
    content = oe2003.create(
        entries=[
            Entry(first_name="Núria", last_name="Pavić", club_id=1, club="OC Kovač"),
            Entry(
                first_name="Núria Istenič",
                last_name="Müller",
                club_id=2,
                club="Futó Club",
            ),
            Entry(
                first_name="Pattantyús", last_name="Hugo", club_id=3, club="OC Ábrahám"
            ),
        ],
        class_list=[],
    )
    v1 = (
        "1;;;Pavic;Núria;;;;0;;;;;1;;OC Kovac;;;;"
        + "\r\n"
        + "2;;;Müller;Nuria Istenic;;;;0;;;;;2;;Futó Club;;;;"
        + "\r\n"
        + "3;;;Hugo;Pattantyús;;;;0;;;;;3;;OC Ábrahám;;;;"
        + "\r\n"
    )

    assert content == bytes(header + "\r\n" + v1, encoding="windows-1252")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ooresults.plugins import iof_class_list


def test_import_class_list():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ClassList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Class>
      <Name>D 14-15</Name>
  </Class>
  <Class>
      <Name>H 14-15</Name>
  </Class>
  <Class>
      <Name>Schüler B</Name>
      <ShortName>SchB</ShortName>
  </Class>
  <Class>
      <Name>Offen Mittel</Name>
      <ShortName>OM</ShortName>
  </Class>
</ClassList>
"""
    classes = iof_class_list.parse_class_list(bytes(content, encoding="utf-8"))
    assert classes == [
        {
            "name": "D 14-15",
            "short_name": None,
        },
        {
            "name": "H 14-15",
            "short_name": None,
        },
        {
            "name": "Schüler B",
            "short_name": "SchB",
        },
        {
            "name": "Offen Mittel",
            "short_name": "OM",
        },
    ]


def test_export_class_list():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<ClassList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Class>
    <Name>D 14-15</Name>
  </Class>
  <Class>
    <Name>H 14-15</Name>
  </Class>
  <Class>
    <Name>Schüler B</Name>
    <ShortName>SchB</ShortName>
  </Class>
  <Class>
    <Name>Offen Mittel</Name>
    <ShortName>OM</ShortName>
  </Class>
</ClassList>
"""
    document = iof_class_list.create_class_list(
        [
            {
                "name": "D 14-15",
                "short_name": None,
            },
            {
                "name": "H 14-15",
            },
            {
                "name": "Schüler B",
                "short_name": "SchB",
            },
            {
                "name": "Offen Mittel",
                "short_name": "OM",
            },
        ]
    )
    assert document == bytes(content, encoding="utf-8")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

import pytest

from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.class_params import ClassParams
from ooresults.handler import model


@pytest.fixture
def db():
    model.db = SqliteRepo(db=":memory:")
    return model.db


@pytest.fixture
def event_id(db):
    return db.add_event(
        name="Event",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=False,
    )


@pytest.fixture
def course_1_id(db, event_id):
    return db.add_course(
        event_id=event_id,
        name="Bahn A",
        length=4500,
        climb=90,
        controls=["101", "102", "103"],
    )


@pytest.fixture
def class_1_id(db, event_id, course_1_id):
    return db.add_class(
        event_id=event_id,
        name="Elite Men",
        short_name="E Men",
        course_id=course_1_id,
        params=ClassParams(),
    )


@pytest.fixture
def class_2_id(db, event_id):
    return db.add_class(
        event_id=event_id,
        name="Elite Women",
        short_name="E Women",
        course_id=None,
        params=ClassParams(),
    )


def test_import_course_data_with_climb_update_existing_course(
    db, event_id, course_1_id
):
    courses = [
        {
            "name": "Bahn A",
            "length": 4800,
            "climb": 120,
            "controls": ["101", "102"],
        },
    ]
    model.import_courses(event_id=event_id, courses=courses, class_course=[])

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 1
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length == 4800
    assert co[0].climb == 120
    assert co[0].controls == ["101", "102"]


def test_import_course_data_without_climb_update_existing_course(
    db, event_id, course_1_id
):
    courses = [
        {
            "name": "Bahn A",
            "length": None,
            "climb": None,
            "controls": ["101", "102"],
        },
    ]
    model.import_courses(event_id=event_id, courses=courses, class_course=[])

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 1
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length is None
    assert co[0].climb is None
    assert co[0].controls == ["101", "102"]


def test_import_course_data_add_not_existing_course(db, event_id, course_1_id):
    courses = [
        {
            "name": "Bahn B",
            "length": 4800,
            "climb": 120,
            "controls": ["104"],
        },
    ]
    model.import_courses(event_id=event_id, courses=courses, class_course=[])

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 2
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length == 4500
    assert co[0].climb == 90
    assert co[0].controls == ["101", "102", "103"]
    assert co[1].id != co[0].id
    assert co[1].name == "Bahn B"
    assert co[1].length == 4800
    assert co[1].climb == 120
    assert co[1].controls == ["104"]


def test_import_course_data_update_or_add_courses(db, event_id, course_1_id):
    courses = [
        {
            "name": "Bahn A",
            "length": 4800,
            "climb": None,
            "controls": ["101", "102"],
        },
        {
            "name": "Bahn B",
            "length": 3900,
            "climb": 120,
            "controls": ["201", "102"],
        },
        {
            "name": "Bahn C",
            "length": None,
            "climb": None,
            "controls": [],
        },
    ]
    model.import_courses(event_id=event_id, courses=courses, class_course=[])

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 3
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length == 4800
    assert co[0].climb is None
    assert co[0].controls == ["101", "102"]
    assert co[1].id != co[0].id
    assert co[1].name == "Bahn B"
    assert co[1].length == 3900
    assert co[1].climb == 120
    assert co[1].controls == ["201", "102"]
    assert co[2].id != co[0].id
    assert co[2].name == "Bahn C"
    assert co[2].length is None
    assert co[2].climb is None
    assert co[2].controls == []


def test_import_course_data_add_not_existing_class(
    db, event_id, class_1_id, course_1_id
):
    class_course = [
        {
            "class_name": "Beginners",
            "course_name": None,
        }
    ]
    model.import_courses(event_id=event_id, courses=[], class_course=class_course)

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 1
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length == 4500
    assert co[0].climb == 90
    assert co[0].controls == ["101", "102", "103"]
    cl = list(model.get_classes(event_id=event_id))
    assert len(cl) == 2
    assert cl[0].id != cl[1].id
    assert cl[0].name == "Beginners"
    assert cl[0].short_name == None
    assert cl[0].course_id == None
    assert cl[0].course == None
    assert cl[0].params == ClassParams()
    assert cl[1].id == class_1_id
    assert cl[1].name == "Elite Men"
    assert cl[1].short_name == "E Men"
    assert cl[1].course_id == course_1_id
    assert cl[1].course == "Bahn A"
    assert cl[1].params == ClassParams()


def test_import_course_data_update_class_course_assignment(
    db, event_id, class_1_id, class_2_id, course_1_id
):
    courses = [
        {
            "name": "Bahn B",
            "length": 3900,
            "climb": 120,
            "controls": ["201", "102"],
        },
    ]
    class_course = [
        {
            "class_name": "Elite Men",
            "course_name": "Bahn B",
        }
    ]
    model.import_courses(event_id=event_id, courses=courses, class_course=class_course)

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 2
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length == 4500
    assert co[0].climb == 90
    assert co[0].controls == ["101", "102", "103"]
    assert co[1].id != course_1_id
    assert co[1].name == "Bahn B"
    assert co[1].length == 3900
    assert co[1].climb == 120
    assert co[1].controls == ["201", "102"]
    cl = list(model.get_classes(event_id=event_id))
    assert len(cl) == 2
    assert cl[0].id == class_1_id
    assert cl[0].name == "Elite Men"
    assert cl[0].short_name == "E Men"
    assert cl[0].course_id == co[1].id
    assert cl[0].course == "Bahn B"
    assert cl[0].params == ClassParams()
    assert cl[1].id == class_2_id
    assert cl[1].name == "Elite Women"
    assert cl[1].short_name == "E Women"
    assert cl[1].course_id is None
    assert cl[1].course is None
    assert cl[1].params == ClassParams()


def test_import_course_data_remove_class_course_assigment(
    db, event_id, class_1_id, course_1_id
):
    class_course = [
        {
            "class_name": "Elite Men",
            "course_name": None,
        }
    ]
    model.import_courses(event_id=event_id, courses=[], class_course=class_course)

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 1
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length == 4500
    assert co[0].climb == 90
    assert co[0].controls == ["101", "102", "103"]
    cl = list(model.get_classes(event_id=event_id))
    assert len(cl) == 1
    assert cl[0].id == class_1_id
    assert cl[0].name == "Elite Men"
    assert cl[0].short_name == "E Men"
    assert cl[0].course_id is None
    assert cl[0].course is None
    assert cl[0].params == ClassParams()


def test_import_course_data_update_or_add_class_course_assigments(
    db, event_id, course_1_id, class_1_id, class_2_id
):
    courses = [
        {
            "name": "Bahn A",
            "length": 4800,
            "climb": 120,
            "controls": ["101", "102"],
        },
        {
            "name": "Bahn B",
            "length": None,
            "climb": None,
            "controls": ["104"],
        },
    ]
    class_course = [
        {
            "class_name": "Elite Men",
            "course_name": "Bahn B",
        },
        {
            "class_name": "Elite Women",
            "course_name": "Bahn A",
        },
        {
            "class_name": "Beginners",
            "course_name": None,
        },
        {
            "class_name": "Elite",
            "course_name": "Bahn B",
        },
    ]
    model.import_courses(event_id=event_id, courses=courses, class_course=class_course)

    co = list(model.get_courses(event_id=event_id))
    assert len(co) == 2
    assert co[0].id == course_1_id
    assert co[0].name == "Bahn A"
    assert co[0].length == 4800
    assert co[0].climb == 120
    assert co[0].controls == ["101", "102"]
    assert co[1].id != co[0].id
    assert co[1].name == "Bahn B"
    assert co[1].length == None
    assert co[1].climb == None
    assert co[1].controls == ["104"]

    cl = list(model.get_classes(event_id=event_id))
    assert len(cl) == 4
    assert cl[0].id not in (cl[2], cl[3])
    assert cl[0].name == "Beginners"
    assert cl[0].course_id == None
    assert cl[0].course == None
    assert cl[0].params == ClassParams()
    assert cl[1].id not in (cl[2], cl[3])
    assert cl[1].name == "Elite"
    assert cl[1].short_name == None
    assert cl[1].course_id == co[1].id
    assert cl[1].course == "Bahn B"
    assert cl[1].params == ClassParams()
    assert cl[2].id == class_1_id
    assert cl[2].name == "Elite Men"
    assert cl[2].short_name == "E Men"
    assert cl[2].course_id == co[1].id
    assert cl[2].course == "Bahn B"
    assert cl[2].params == ClassParams()
    assert cl[3].id == class_2_id
    assert cl[3].name == "Elite Women"
    assert cl[3].short_name == "E Women"
    assert cl[3].course_id == course_1_id
    assert cl[3].course == "Bahn A"
    assert cl[3].params == ClassParams()



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from datetime import datetime
from datetime import timezone

import pytest

from ooresults.handler import handicap
from ooresults.repo.result_type import SplitTime
from ooresults.repo.result_type import PersonRaceResult
from ooresults.repo.result_type import ResultStatus
from ooresults.repo.class_params import ClassParams


def t(a: datetime, b: datetime) -> int:
    diff = b.replace(microsecond=0) - a.replace(microsecond=0)
    return int(diff.total_seconds())


@pytest.mark.parametrize("time_limit, score_overtime", [(60, -2), (120, -1), (180, 0)])
def test_compute_result_status_ok(time_limit, score_overtime):
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 3, tzinfo=timezone.utc)
    c3 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 41, 7, tzinfo=timezone.utc)

    controls = ["101", "102", "103"]
    result = PersonRaceResult(
        status=ResultStatus.INACTIVE,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="103", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="101", punch_time=c3, status="Additional"),
        ],
    )
    class_params = ClassParams(otype="score", time_limit=time_limit)

    result.compute_result(controls=controls, class_params=class_params)
    assert result.start_time == s1
    assert result.punched_start_time == s1
    assert result.finish_time == f1
    assert result.punched_finish_time == f1
    assert result.time == int((f1 - s1).total_seconds())
    assert result.status == ResultStatus.OK
    assert len(result.split_times) == 3
    assert result.split_times[0] == SplitTime(
        control_code="103", punch_time=c1, time=t(s1, c1), status="OK"
    )
    assert result.split_times[1] == SplitTime(
        control_code="102", punch_time=c2, time=t(s1, c2), status="OK"
    )
    assert result.split_times[2] == SplitTime(
        control_code="101", punch_time=c3, time=t(s1, c3), status="OK"
    )
    assert len(result.extensions) == 3
    assert result.extensions["score_controls"] == 3
    assert result.extensions["score_overtime"] == score_overtime
    assert result.extensions["score"] == 3 + score_overtime


def test_compute_result_status_mispunched():
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c3 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    controls = ["101", "102", "103", "104"]
    result = PersonRaceResult(
        status=ResultStatus.INACTIVE,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="103", punch_time=c3, status="Additional"),
        ],
    )
    class_params = ClassParams(otype="score", time_limit=60)

    result.compute_result(controls=controls, class_params=class_params)
    assert result.start_time == s1
    assert result.punched_start_time == s1
    assert result.finish_time == f1
    assert result.punched_finish_time == f1
    assert result.time == int((f1 - s1).total_seconds())
    assert result.status == ResultStatus.OK
    assert len(result.split_times) == 4
    assert result.split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=t(s1, c1), status="OK"
    )
    assert result.split_times[1] == SplitTime(
        control_code="103", punch_time=c3, time=t(s1, c3), status="OK"
    )
    assert result.split_times[2] == SplitTime(
        control_code="102", punch_time=None, time=None, status="Missing"
    )
    assert result.split_times[3] == SplitTime(
        control_code="104", punch_time=None, time=None, status="Missing"
    )
    assert len(result.extensions) == 3
    assert result.extensions["score_controls"] == 2
    assert result.extensions["score_overtime"] == 0
    assert result.extensions["score"] == 2


def test_compute_result_status_ok_with_additionals():
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 3, tzinfo=timezone.utc)
    c3 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    c4 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)
    c5 = datetime(2015, 1, 1, 12, 39, 9, tzinfo=timezone.utc)
    c6 = datetime(2015, 1, 1, 12, 39, 11, tzinfo=timezone.utc)
    c7 = datetime(2015, 1, 1, 12, 39, 13, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 15, tzinfo=timezone.utc)

    controls = ["101", "102", "103"]
    result = PersonRaceResult(
        status=ResultStatus.INACTIVE,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="105", punch_time=c2, status="Additional"),
            SplitTime(control_code="101", punch_time=c3, status="Additional"),
            SplitTime(control_code="103", punch_time=c4, status="Additional"),
            SplitTime(control_code="102", punch_time=c5, status="Additional"),
            SplitTime(control_code="101", punch_time=c6, status="Additional"),
            SplitTime(control_code="104", punch_time=c7, status="Additional"),
        ],
    )
    class_params = ClassParams(otype="score", time_limit=60)

    result.compute_result(controls=controls, class_params=class_params)
    assert result.start_time == s1
    assert result.punched_start_time == s1
    assert result.finish_time == f1
    assert result.punched_finish_time == f1
    assert result.time == int((f1 - s1).total_seconds())
    assert result.status == ResultStatus.OK
    assert len(result.split_times) == 7
    assert result.split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=t(s1, c1), status="OK"
    )
    assert result.split_times[1] == SplitTime(
        control_code="105", punch_time=c2, time=t(s1, c2), status="Additional"
    )
    assert result.split_times[2] == SplitTime(
        control_code="101", punch_time=c3, time=t(s1, c3), status="Additional"
    )
    assert result.split_times[3] == SplitTime(
        control_code="103", punch_time=c4, time=t(s1, c4), status="OK"
    )
    assert result.split_times[4] == SplitTime(
        control_code="102", punch_time=c5, time=t(s1, c5), status="OK"
    )
    assert result.split_times[5] == SplitTime(
        control_code="101", punch_time=c6, time=t(s1, c6), status="Additional"
    )
    assert result.split_times[6] == SplitTime(
        control_code="104", punch_time=c7, time=t(s1, c7), status="Additional"
    )
    assert len(result.extensions) == 3
    assert result.extensions["score_controls"] == 3
    assert result.extensions["score_overtime"] == 0
    assert result.extensions["score"] == 3


def test_compute_result_status_last_three_stations_missing():
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 3, tzinfo=timezone.utc)
    c3 = datetime(2015, 1, 1, 12, 39, 11, tzinfo=timezone.utc)
    c4 = datetime(2015, 1, 1, 12, 39, 13, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 15, tzinfo=timezone.utc)

    controls = ["101", "102", "103", "104", "105"]
    result = PersonRaceResult(
        status=ResultStatus.INACTIVE,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="106", punch_time=c3, status="Additional"),
            SplitTime(control_code="107", punch_time=c4, status="Additional"),
        ],
    )
    class_params = ClassParams(otype="score", time_limit=60)

    result.compute_result(controls=controls, class_params=class_params)
    assert result.start_time == s1
    assert result.punched_start_time == s1
    assert result.finish_time == f1
    assert result.punched_finish_time == f1
    assert result.time == int((f1 - s1).total_seconds())
    assert result.status == ResultStatus.OK
    assert len(result.split_times) == 7
    assert result.split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=t(s1, c1), status="OK"
    )
    assert result.split_times[1] == SplitTime(
        control_code="102", punch_time=c2, time=t(s1, c2), status="OK"
    )
    assert result.split_times[2] == SplitTime(
        control_code="106", punch_time=c3, time=t(s1, c3), status="Additional"
    )
    assert result.split_times[3] == SplitTime(
        control_code="107", punch_time=c4, time=t(s1, c4), status="Additional"
    )
    assert result.split_times[4] == SplitTime(
        control_code="103", punch_time=None, time=None, status="Missing"
    )
    assert result.split_times[5] == SplitTime(
        control_code="104", punch_time=None, time=None, status="Missing"
    )
    assert result.split_times[6] == SplitTime(
        control_code="105", punch_time=None, time=None, status="Missing"
    )
    assert len(result.extensions) == 3
    assert result.extensions["score_controls"] == 2
    assert result.extensions["score_overtime"] == 0
    assert result.extensions["score"] == 2


def test_given_no_controls_but_punches_when_compute_results_then_status_is_ok():
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 3, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)

    controls = []
    result = PersonRaceResult(
        status=ResultStatus.INACTIVE,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
        ],
    )
    class_params = ClassParams(otype="score", time_limit=60)

    result.compute_result(controls=controls, class_params=class_params)
    assert result.start_time == s1
    assert result.punched_start_time == s1
    assert result.finish_time == f1
    assert result.punched_finish_time == f1
    assert result.time == int((f1 - s1).total_seconds())
    assert result.status == ResultStatus.OK
    assert len(result.split_times) == 2
    assert result.split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=t(s1, c1), status="Additional"
    )
    assert result.split_times[1] == SplitTime(
        control_code="102", punch_time=c2, time=t(s1, c2), status="Additional"
    )
    assert len(result.extensions) == 3
    assert result.extensions["score_controls"] == 0
    assert result.extensions["score_overtime"] == 0
    assert result.extensions["score"] == 0


@pytest.mark.parametrize("time_limit, score_overtime", [(60, -2), (120, -1), (180, 0)])
@pytest.mark.parametrize(
    "female, year_of_birth", [(True, 2000), (True, 1981), (False, 1941)]
)
def test_compute_handicap_ok(time_limit, score_overtime, female, year_of_birth):
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 41, 7, tzinfo=timezone.utc)

    controls = ["101"]
    result = PersonRaceResult(
        status=ResultStatus.INACTIVE,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
        ],
    )
    class_params = ClassParams(
        otype="score", time_limit=time_limit, apply_handicap_rule=True
    )

    result.compute_result(
        controls=controls,
        class_params=class_params,
        year=year_of_birth,
        gender="F" if female else "M",
    )

    h = handicap.Handicap()
    assert result.start_time == s1
    assert result.punched_start_time == s1
    assert result.finish_time == f1
    assert result.punched_finish_time == f1
    assert result.extensions["factor"] == h.factor(
        female=female, year=f1.year - year_of_birth
    )
    assert result.time == int((f1 - s1).total_seconds())
    assert result.status == ResultStatus.OK
    assert len(result.split_times) == 1
    assert result.split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=t(s1, c1), status="OK"
    )
    assert len(result.extensions) == 4
    assert result.extensions["score_controls"] == 1
    assert result.extensions["score_overtime"] == score_overtime
    assert (
        result.extensions["score"] == 1 / result.extensions["factor"] + score_overtime
    )


@pytest.mark.parametrize(
    "female, year_of_birth", [(True, 2000), (True, 1981), (False, 1941)]
)
def test_compute_handicap_mp(female, year_of_birth):
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    controls = ["101"]
    result = PersonRaceResult(
        status=ResultStatus.INACTIVE,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[],
    )
    class_params = ClassParams(otype="score", time_limit=60, apply_handicap_rule=True)

    result.compute_result(
        controls=controls,
        class_params=class_params,
        year=year_of_birth,
        gender="F" if female else "M",
    )

    h = handicap.Handicap()
    assert result.start_time == s1
    assert result.punched_start_time == s1
    assert result.finish_time == f1
    assert result.punched_finish_time == f1
    assert result.extensions["factor"] == h.factor(
        female=female, year=f1.year - year_of_birth
    )
    assert result.time == int((f1 - s1).total_seconds())
    assert result.status == ResultStatus.OK
    assert len(result.split_times) == 1
    assert result.split_times[0] == SplitTime(
        control_code="101", punch_time=None, time=None, status="Missing"
    )
    assert len(result.extensions) == 4
    assert result.extensions["score_controls"] == 0
    assert result.extensions["score_overtime"] == 0
    assert result.extensions["score"] == 0



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ooresults.repo.result_type import ResultStatus


MAP_STATUS = {
    ResultStatus.INACTIVE: "",
    ResultStatus.FINISHED: "Finished",
    ResultStatus.OK: "OK",
    ResultStatus.MISSING_PUNCH: "MP",
    ResultStatus.DID_NOT_START: "DNS",
    ResultStatus.DID_NOT_FINISH: "DNF",
    ResultStatus.OVER_TIME: "OTL",
    ResultStatus.DISQUALIFIED: "DSQ",
}


EXPERIMENTAL = False


t_globals = {
    "str": str,
    "ResultStatus": ResultStatus,
    "MAP_STATUS": MAP_STATUS,
    "EXPERIMENTAL": EXPERIMENTAL,
}



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import datetime
import pathlib
import json
from typing import Optional
from typing import List
from typing import Dict
from typing import Tuple

import iso8601
import jsonschema

from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.repo import EventNotFoundError
from ooresults.repo.repo import TransactionMode
from ooresults.repo.class_params import ClassParams
from ooresults.repo import result_type
from ooresults.repo import start_type
from ooresults.repo.result_type import ResultStatus
from ooresults.repo.series_type import Settings
from ooresults.websocket_server.websocket_server import WebSocketServer


websocket_server: Optional[WebSocketServer] = None
db: Optional[SqliteRepo] = None


data_path = (
    pathlib.Path(__file__).resolve().parent.parent / "schema" / "cardreader_log.json"
)
with open(data_path, "r") as f:
    schema_cardreader_log = json.loads(f.read())


def get_classes(event_id):
    with db.transaction():
        return db.get_classes(event_id=event_id)


def get_class(id):
    with db.transaction():
        return db.get_class(id=id)


def import_classes(event_id: int, classes: List[Dict]) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        existing_classes = list(db.get_classes(event_id=event_id))
        for c in classes:
            for cl in existing_classes:
                if cl["name"] == c["name"]:
                    db.update_class(
                        id=cl["id"],
                        name=cl["name"],
                        short_name=c["short_name"]
                        if c.get("short_name", None) is not None
                        else cl["short_name"],
                        course_id=cl["course_id"],
                        params=cl.params,
                    )
                    break
            else:
                id = db.add_class(
                    event_id=event_id,
                    name=c["name"],
                    short_name=c.get("short_name", None),
                    course_id=None,
                    params=ClassParams(),
                )


def add_class(
    event_id,
    name,
    short_name: Optional[str],
    course_id: Optional[int],
    params: ClassParams,
):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.add_class(
            event_id=event_id,
            name=name,
            short_name=short_name,
            course_id=course_id,
            params=params,
        )


def update_class(
    id: int,
    event_id: int,
    name: str,
    short_name: Optional[str],
    course_id: Optional[int],
    params: ClassParams,
) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        # update class data
        try:
            db.update_class(
                id=id,
                name=name,
                short_name=short_name,
                course_id=course_id,
                params=params,
            )
        except KeyError:
            # distinguish between 'Event deleted' or 'Class deleted'
            db.get_event(id=event_id)
            raise

        # update results of the competitors belonging to the modified class
        try:
            class_params = params
            controls = db.get_course(id=course_id)[0]["controls"]
        except:
            class_params = ClassParams()
            controls = []

        entries = db.get_entries(event_id=event_id)
        for entry in entries:
            if entry.class_id == id:
                entry.result.compute_result(
                    controls=controls,
                    class_params=class_params,
                    start_time=entry.start.start_time,
                    year=entry.year,
                    gender=entry.gender if entry.gender != "" else None,
                )
                db.update_entry_result(
                    id=entry["id"],
                    chip=entry["chip"],
                    result=entry.result,
                    start_time=entry.start.start_time,
                )


def delete_classes(event_id):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_classes(event_id=event_id)


def delete_class(id):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_class(id=id)


def get_courses(event_id: int):
    with db.transaction():
        return db.get_courses(event_id=event_id)


def get_course(id):
    with db.transaction():
        return db.get_course(id=id)


def import_courses(
    event_id: int, courses: List[Dict], class_course: List[Dict]
) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        existing_courses = list(db.get_courses(event_id=event_id))
        for c in courses:
            for co in existing_courses:
                if co["name"] == c["name"]:
                    db.update_course(
                        id=co["id"],
                        name=co["name"],
                        length=c.get("length", None),
                        climb=c.get("climb", None),
                        controls=c["controls"],
                    )
                    break
            else:
                id = db.add_course(
                    event_id=event_id,
                    name=c["name"],
                    length=c.get("length", None),
                    climb=c.get("climb", None),
                    controls=c["controls"],
                )
        existing_classes = list(db.get_classes(event_id=event_id))
        existing_courses = list(db.get_courses(event_id=event_id))
        for i in class_course:
            for co in existing_courses:
                if co["name"] == i.get("course_name", None):
                    for cl in existing_classes:
                        if cl["name"] == i["class_name"]:
                            db.update_class(
                                id=cl["id"],
                                name=cl["name"],
                                short_name=cl["short_name"],
                                course_id=co["id"],
                                params=cl.params,
                            )
                            break
                    else:
                        id = db.add_class(
                            event_id=event_id,
                            name=i["class_name"],
                            short_name=None,
                            course_id=co["id"],
                            params=ClassParams(),
                        )
                        existing_classes.append({"id": id, "name": i["class_name"]})
                    break
            else:
                if i["course_name"] is None:
                    for cl in existing_classes:
                        if cl["name"] == i["class_name"]:
                            db.update_class(
                                id=cl["id"],
                                name=cl["name"],
                                short_name=cl["short_name"],
                                course_id=None,
                                params=cl.params,
                            )
                            break
                    else:
                        id = db.add_class(
                            event_id=event_id,
                            name=i["class_name"],
                            short_name=None,
                            course_id=None,
                            params=ClassParams(),
                        )
                        existing_classes.append({"id": id, "name": i["class_name"]})


def add_course(
    event_id: int,
    name: str,
    length: Optional[float],
    climb: Optional[float],
    controls: List[str],
):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.add_course(
            event_id=event_id,
            name=name,
            length=length,
            climb=climb,
            controls=controls,
        )


def update_course(
    id: int,
    event_id: int,
    name: str,
    length: Optional[float],
    climb: Optional[float],
    controls: List[str],
) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        # update course data
        try:
            db.update_course(
                id=id,
                name=name,
                length=length,
                climb=climb,
                controls=controls,
            )
        except KeyError:
            # distinguish between 'Event deleted' or 'Course deleted'
            db.get_event(id=event_id)
            raise

        # update results of the competitors belonging to a class using the modified course
        classes = db.get_classes(event_id=event_id)
        classes = [c for c in classes if c.course_id == id]

        entries = db.get_entries(event_id=event_id)
        for entry in entries:
            for class_ in classes:
                if entry.class_id == class_.id:
                    entry.result.compute_result(
                        controls=controls,
                        class_params=class_.params,
                        start_time=entry.start.start_time,
                        year=entry.year,
                        gender=entry.gender if entry.gender != "" else None,
                    )
                    db.update_entry_result(
                        id=entry["id"],
                        chip=entry["chip"],
                        result=entry.result,
                        start_time=entry.start.start_time,
                    )
                    break


def delete_courses(event_id):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_courses(event_id=event_id)


def delete_course(id):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_course(id=id)


def get_clubs():
    with db.transaction():
        return db.get_clubs()


def get_club(id):
    with db.transaction():
        return db.get_club(id=id)


def add_club(name):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.add_club(name=name)


def update_club(id, name):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.update_club(id=id, name=name)


def delete_club(id):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_club(id=id)


def get_competitors():
    with db.transaction():
        return db.get_competitors()


def get_competitor(id):
    with db.transaction():
        return db.get_competitor(id=id)


def add_competitor(first_name, last_name, club_id, gender, year, chip):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.add_competitor(
            first_name=first_name,
            last_name=last_name,
            club_id=club_id,
            gender=gender,
            year=year,
            chip=chip,
        )


def update_competitor(id, first_name, last_name, club_id, gender, year, chip):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.update_competitor(
            id=id,
            first_name=first_name,
            last_name=last_name,
            club_id=club_id,
            gender=gender,
            year=year,
            chip=chip,
        )


def delete_competitor(id):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_competitor(id=id)


def import_competitors(competitors):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.import_competitors(competitors)


def import_entries(event_id, entries):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.import_entries(event_id=event_id, entries=entries)


def get_entries(event_id):
    with db.transaction():
        return db.get_entries(event_id=event_id)


def get_entry(id):
    with db.transaction():
        return db.get_entry(id=id)


def add_entry(
    event_id: int,
    competitor_id: Optional[int],
    first_name: str,
    last_name: str,
    gender: str,
    year: Optional[int],
    class_id: int,
    club_id: Optional[int],
    not_competing: bool,
    chip: str,
    fields: Dict[int, str],
    status: ResultStatus,
    start_time: Optional[datetime.datetime],
) -> int:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        return db.add_entry(
            event_id=event_id,
            competitor_id=competitor_id,
            first_name=first_name,
            last_name=last_name,
            gender=gender,
            year=year,
            class_id=class_id,
            club_id=club_id,
            not_competing=not_competing,
            chip=chip,
            fields=fields,
            status=status,
            start_time=start_time,
        )


def update_entry(
    id: int,
    first_name: str,
    last_name: str,
    gender: str,
    year: Optional[int],
    class_id: int,
    club_id: Optional[int],
    not_competing: bool,
    chip: str,
    fields: Dict[int, str],
    status: ResultStatus,
    start_time: Optional[datetime.datetime],
) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.update_entry(
            id=id,
            first_name=first_name,
            last_name=last_name,
            gender=gender,
            year=year,
            class_id=class_id,
            club_id=club_id,
            not_competing=not_competing,
            chip=chip,
            fields=fields,
            status=status,
            start_time=start_time,
        )


def add_entry_result(event_id, chip, start_time, result):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.add_entry_result(
            event_id=event_id,
            chip=chip,
            start_time=start_time,
            result=result,
        )


def update_entry_result(id, chip, start_time, result):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.update_entry_result(
            id=id,
            chip=chip,
            start_time=start_time,
            result=result,
        )


def delete_entries(event_id: int):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_entries(event_id=event_id)


def delete_entry(id: int):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_entry(id=id)


def get_events():
    with db.transaction():
        return db.get_events()


def get_event(id: int):
    with db.transaction():
        return db.get_event(id=id)


def add_event(
    name: str,
    date: datetime.date,
    key: Optional[str],
    publish: bool,
    series: Optional[str],
    fields: List[str] = [],
) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.add_event(name, date, key, publish, series, fields)


def update_event(
    id: int,
    name: str,
    date: datetime.date,
    key: Optional[str],
    publish: bool,
    series: Optional[str],
    fields: List[str] = [],
) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.update_event(id, name, date, key, publish, series, fields)

    future = asyncio.run_coroutine_threadsafe(
        coro=websocket_server.handler.update_event(
            {"id": id, "name": name, "date": date}
        ),
        loop=websocket_server.loop,
    )
    future.result()


def delete_event(id):
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.delete_entries(event_id=id)
        db.delete_classes(event_id=id)
        db.delete_courses(event_id=id)
        db.delete_event(id)


def parse_cardreader_log(item: Dict) -> Dict:
    #
    # returns a dict consisting of:
    #
    # 'entryType': str,
    # 'entryTime': datetime.dateime,
    # 'controlCard': Optional[str],
    # 'result' : Optional[result_type.PersonRaceResult]
    #
    jsonschema.validate(item, schema_cardreader_log)
    d = {
        "entryType": item["entryType"],
        "entryTime": iso8601.parse_date(item["entryTime"]),
        "controlCard": item.get("controlCard", None),
        "result": None,
    }

    if d["entryType"] == "cardRead":
        result = result_type.PersonRaceResult(status=ResultStatus.FINISHED)
        if item.get("clearTime", None) is not None:
            result.punched_clear_time = iso8601.parse_date(item["clearTime"])
        if item.get("checkTime", None) is not None:
            result.punched_check_time = iso8601.parse_date(item["checkTime"])
        if item.get("startTime", None) is not None:
            result.punched_start_time = iso8601.parse_date(item["startTime"])
        if item.get("finishTime", None) is not None:
            result.punched_finish_time = iso8601.parse_date(item["finishTime"])
        result.start_time = result.punched_start_time
        result.finish_time = result.punched_finish_time
        for p in item["punches"]:
            result.split_times.append(
                result_type.SplitTime(
                    control_code=p["controlCode"],
                    punch_time=iso8601.parse_date(p["punchTime"]),
                    status="Additional",
                )
            )
        d["result"] = result

    return d


def store_cardreader_result(event_key: str, item: Dict) -> Tuple[str, Dict, Dict]:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        for e in db.get_events():
            if event_key != "" and e.key == event_key:
                event = e
                break
        else:
            raise EventNotFoundError(f'Event for key "{event_key}" not found')

        if item["entryType"] == "cardRead":
            entries = db.get_entries(event_id=event.id)
            entries_controlcard = [
                e
                for e in entries
                if e.class_id is not None and e.chip == item["controlCard"]
            ]
            result = item["result"]

            if len(entries_controlcard) == 1:
                entry = entries_controlcard[0]
                try:
                    class_ = db.get_class(id=entry["class_id"])[0]
                    course_id = class_["course_id"]
                    class_params = class_["params"]
                    controls = db.get_course(id=course_id)[0]["controls"]
                except:
                    class_params = ClassParams()
                    controls = []

                result.compute_result(
                    controls=controls,
                    class_params=class_params,
                    start_time=entry.start.start_time,
                    year=int(entry.get("year", None))
                    if entry.get("year", None) is not None
                    else None,
                    gender=entry.get("gender", None),
                )

                # do not update another existing result
                if not entry.result.has_punches():
                    db.update_entry_result(
                        id=entry["id"],
                        chip=entry["chip"],
                        result=result,
                        start_time=entry["start"].start_time,
                    )
                    res = {
                        "entryTime": item["entryTime"],
                        "eventId": event.id,
                        "controlCard": entry["chip"],
                        "firstName": entry["first_name"],
                        "lastName": entry["last_name"],
                        "club": entry["club"],
                        "class": entry["class_"],
                        "status": result["status"],
                        "time": result.extensions.get("running_time", result["time"]),
                        "error": None,
                    }
                else:
                    # do not create a new entry if an entry with same result already exist
                    if entry.result == result:
                        res = {
                            "entryTime": item["entryTime"],
                            "eventId": event.id,
                            "controlCard": entry["chip"],
                            "firstName": entry["first_name"],
                            "lastName": entry["last_name"],
                            "club": entry["club"],
                            "class": entry["class_"],
                            "status": result["status"],
                            "time": result.extensions.get(
                                "running_time", result["time"]
                            ),
                            "error": None,
                        }
                    else:
                        result.status = ResultStatus.FINISHED
                        result.compute_result(controls=[], class_params=ClassParams())
                        db.add_entry_result(
                            event_id=event.id,
                            chip=item["controlCard"],
                            result=result,
                            start_time=None,
                        )
                        res = {
                            "entryTime": item["entryTime"],
                            "eventId": event.id,
                            "controlCard": item["controlCard"],
                            "firstName": None,
                            "lastName": None,
                            "club": None,
                            "class": None,
                            "status": result["status"],
                            "time": None,
                            "error": "There is already a result",
                        }

            else:
                result.compute_result(controls=[], class_params=ClassParams())
                # do not create a new entry if an entry with same results already exist
                for entry in [
                    e
                    for e in entries
                    if e.class_id is None and e.chip == item["controlCard"]
                ]:
                    if entry.result == result:
                        res = {
                            "entryTime": item["entryTime"],
                            "eventId": event.id,
                            "controlCard": entry["chip"],
                            "firstName": entry["first_name"],
                            "lastName": entry["last_name"],
                            "club": entry["club"],
                            "class": entry["class_"],
                            "status": result["status"],
                            "time": result.extensions.get(
                                "running_time", result["time"]
                            ),
                        }
                        break
                else:
                    db.add_entry_result(
                        event_id=event.id,
                        chip=item["controlCard"],
                        result=result,
                        start_time=None,
                    )
                    res = {
                        "entryTime": item["entryTime"],
                        "eventId": event.id,
                        "controlCard": item["controlCard"],
                        "firstName": None,
                        "lastName": None,
                        "club": None,
                        "class": None,
                        "status": result["status"],
                        "time": None,
                    }
                if entries_controlcard == []:
                    res["error"] = "Control card unknown"
                else:
                    res["error"] = "Control card assigned several times"

        elif item["entryType"] == "cardInserted":
            res = {"eventId": event.id, "controlCard": item["controlCard"]}
        else:
            res = {"eventId": event.id}

    return item["entryType"], event, res


def add_or_update_entry(
    id: Optional[int],
    event_id: int,
    competitor_id: Optional[int],
    first_name: str,
    last_name: str,
    gender: str,
    year: Optional[int],
    class_id: int,
    club_id: Optional[int],
    not_competing: bool,
    chip: str,
    fields: Dict[int, str],
    status: ResultStatus,
    start_time: Optional[datetime.datetime],
    result_id: Optional[int],
) -> None:
    #
    # result_id == -1: remove result from entry (store as pseudy result)
    #

    with db.transaction(mode=TransactionMode.IMMEDIATE):

        def store_result_as_new_entry(entry) -> None:
            result = entry.result
            result.status = ResultStatus.FINISHED
            result.split_times = [
                s for s in result.split_times if s.status != "Missing"
            ]
            for s in result.split_times:
                s.status = "Additional"
                if result.start_time is None:
                    s.time = None
            db.add_entry_result(
                event_id=entry.event_id,
                chip=entry.chip,
                result=result,
                start_time=None,
            )

        if id is None:
            entry = {
                "result": result_type.PersonRaceResult(),
                "start": start_type.PersonRaceStart(),
            }
            entry["id"] = db.add_entry(
                event_id=event_id,
                competitor_id=competitor_id,
                first_name=first_name,
                last_name=last_name,
                gender=gender,
                year=year,
                class_id=class_id,
                club_id=club_id,
                not_competing=not_competing,
                chip=chip,
                fields=fields,
                status=status,
                start_time=start_time,
            )
        else:
            entry = db.get_entry(id=id)[0]
            if result_id is not None and entry.result.has_punches():
                store_result_as_new_entry(entry=entry)

            db.update_entry(
                id=id,
                first_name=first_name,
                last_name=last_name,
                gender=gender,
                year=year,
                class_id=class_id,
                club_id=club_id,
                not_competing=not_competing,
                chip=chip,
                fields=fields,
                status=status,
                start_time=start_time,
            )
            if result_id == -1:
                # remove result
                db.update_entry_result(
                    id=id,
                    chip=chip,
                    result=result_type.PersonRaceResult(),
                    start_time=start_time,
                )

        if result_id is not None and result_id != -1:
            # compute new result
            result_entry = db.get_entry(id=result_id)[0]
            try:
                class_ = db.get_class(id=class_id)[0]
                course_id = class_["course_id"]
                class_params = class_["params"]
                controls = db.get_course(id=course_id)[0]["controls"]
            except:
                class_params = ClassParams()
                controls = []

            result_entry.result.compute_result(
                controls=controls,
                class_params=class_params,
                start_time=start_time,
                year=year,
                gender=gender if gender != "" else None,
            )
            db.update_entry_result(
                id=entry["id"],
                chip=result_entry["chip"],
                result=result_entry.result,
                start_time=start_time,
            )
            db.delete_entry(id=result_entry["id"])

        elif id is not None and (
            class_id != entry.class_ or start_time != entry["start"].start_time
        ):
            # compute new result
            result_entry = db.get_entry(id=id)[0]
            if result_entry.result.has_punches():
                try:
                    class_ = db.get_class(id=class_id)[0]
                    course_id = class_["course_id"]
                    class_params = class_["params"]
                    controls = db.get_course(id=course_id)[0]["controls"]
                except:
                    class_params = ClassParams()
                    controls = []

                result_entry.result.compute_result(
                    controls=controls,
                    class_params=class_params,
                    start_time=start_time,
                    year=year,
                    gender=gender if gender != "" else None,
                )
                db.update_entry_result(
                    id=entry["id"],
                    chip=result_entry["chip"],
                    result=result_entry.result,
                    start_time=start_time,
                )


def get_series_settings() -> Settings:
    with db.transaction():
        return db.get_series_settings()


def update_series_settings(settings: Settings) -> None:
    with db.transaction(mode=TransactionMode.IMMEDIATE):
        db.update_series_settings(settings=settings)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import threading
import ssl

import websockets

from ooresults.websocket_server.websocket_handler import WebSocketHandler


class WebSocketServer(threading.Thread):
    def __init__(
        self,
        handler: WebSocketHandler,
        host: str = "0.0.0.0",
        port: int = 8081,
        ssl_cert=None,
        ssl_key=None,
    ):
        super().__init__()
        self.daemon = True
        self.handler = handler
        self.host = host
        self.port = port
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key
        self.loop = None

    def run(self):
        if self.ssl_cert is None:
            ssl_context = None
        else:
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            ssl_context.load_cert_chain(certfile=self.ssl_cert, keyfile=self.ssl_key)

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop=self.loop)
        start_server = websockets.serve(
            self.handler.handler, self.host, self.port, loop=self.loop, ssl=ssl_context
        )
        self.loop.run_until_complete(start_server)
        if ssl_context is None:
            print(f"ws://{self.host}:{str(self.port)}")
        else:
            print(f"wss://{self.host}:{str(self.port)}")
        self.loop.run_forever()
        print("thread finished")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import pathlib

import web
from web.utils import Storage

from ooresults.handler import model
from ooresults.plugins import iof_competitor_list
from ooresults.repo.repo import CompetitorUsedError
from ooresults.repo.repo import ConstraintError
from ooresults.utils.globals import t_globals


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals=t_globals)


class Update:
    def POST(self):
        """Update data"""
        return render.competitors_table(model.get_competitors())


class Add:
    def POST(self):
        """Add or edit entry"""
        data = web.input()
        print(data)
        try:
            if data.id == "":
                model.add_competitor(
                    first_name=data.first_name,
                    last_name=data.last_name,
                    club_id=int(data.club) if data.club != "" else None,
                    gender=data.gender,
                    year=int(data.year) if data.year != "" else None,
                    chip=data.chip,
                )
            else:
                model.update_competitor(
                    id=int(data.id),
                    first_name=data.first_name,
                    last_name=data.last_name,
                    club_id=int(data.club) if data.club != "" else None,
                    gender=data.gender,
                    year=int(data.year) if data.year != "" else None,
                    chip=data.chip,
                )
        except ConstraintError as e:
            raise web.conflict(str(e))
        except:
            logging.exception("Internal server error")
            raise

        return render.competitors_table(model.get_competitors())


class Import:
    def POST(self):
        """Import entries"""
        data = web.input()
        try:
            if data.comp_import == "comp.import.1":
                competitors = iof_competitor_list.parse_competitor_list(
                    content=data.browse1
                )
                model.import_competitors(competitors=competitors)

        except Exception as e:
            raise web.Conflict(str(e))

        return render.competitors_table(model.get_competitors())


class Export:
    def POST(self):
        """Export entries"""
        data = web.input()
        print("###", data)
        try:
            if data.comp_export == "comp.export.1":
                competitors = model.get_competitors()
                content = iof_competitor_list.create_competitor_list(competitors)

        except:
            logging.exception("Internal server error")
            raise

        return content


class Delete:
    def POST(self):
        """Delete entry"""
        data = web.input()
        try:
            model.delete_competitor(int(data.id))
            return render.competitors_table(model.get_competitors())
        except CompetitorUsedError:
            raise web.conflict("Competitor used in entries")
        except:
            logging.exception("Internal server error")
            raise


class FillEditForm:
    def POST(self):
        """Query data to fill add or edit form"""
        data = web.input()
        if data.id == "":
            competitor = Storage(
                {
                    "id": "",
                    "first_name": "",
                    "last_name": "",
                    "club_id": "",
                    "club_name": "",
                    "gender": "",
                    "year": "",
                    "chip": "",
                }
            )
        else:
            competitor = model.get_competitor(int(data.id))[0]

        clubs = model.get_clubs()
        return render.add_competitor(competitor, clubs)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

import pytest

from ooresults.repo import repo
from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.class_params import ClassParams
from ooresults.repo.result_type import ResultStatus


@pytest.fixture
def db():
    return SqliteRepo(db=":memory:")


@pytest.fixture
def club_id(db):
    return db.add_club(
        name="OL Bundestag",
    )


@pytest.fixture
def event_id(db):
    return db.add_event(
        name="Event",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=None,
    )


@pytest.fixture
def class_id(db, event_id):
    return db.add_class(
        event_id=event_id,
        name="Class",
        short_name=None,
        course_id=None,
        params=ClassParams(),
    )


@pytest.fixture
def entry_id(db, event_id, class_id, competitor_1_id):
    return db.add_entry(
        event_id=event_id,
        competitor_id=None,
        first_name="Jogi",
        last_name="Löw",
        gender="M",
        year=None,
        class_id=class_id,
        club_id=None,
        not_competing=False,
        chip="",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )


@pytest.fixture
def competitor_1_id(db, club_id):
    return db.add_competitor(
        first_name="Jogi",
        last_name="Löw",
        club_id=None,
        gender="M",
        year=None,
        chip="",
    )


@pytest.fixture
def competitor_2_id(db, club_id):
    return db.add_competitor(
        first_name="Angela",
        last_name="Merkel",
        club_id=club_id,
        gender="F",
        year=1957,
        chip="1234567",
    )


def test_get_competitors_after_adding_one_competitor(db, competitor_1_id, club_id):
    c = list(db.get_competitors())
    assert len(c) == 1
    assert c[0].id == competitor_1_id
    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id == None
    assert c[0].club_name == None
    assert c[0].gender == "M"
    assert c[0].year is None
    assert c[0].chip == ""


def test_get_competitors_after_adding_two_competitors(
    db, competitor_1_id, competitor_2_id, club_id
):
    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == competitor_1_id
    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id == None
    assert c[0].club_name == None
    assert c[0].gender == "M"
    assert c[0].year is None
    assert c[0].chip == ""
    assert c[1].id == competitor_2_id
    assert c[1].first_name == "Angela"
    assert c[1].last_name == "Merkel"
    assert c[1].club_id == club_id
    assert c[1].club_name == "OL Bundestag"
    assert c[1].gender == "F"
    assert c[1].year == 1957
    assert c[1].chip == "1234567"


def test_get_first_added_competitor(db, competitor_1_id, competitor_2_id, club_id):
    c = list(db.get_competitor(id=competitor_1_id))
    assert len(c) == 1
    assert c[0].id == competitor_1_id
    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id == None
    assert c[0].gender == "M"
    assert c[0].year is None
    assert c[0].chip == ""


def test_get_last_added_competitor(db, competitor_1_id, competitor_2_id, club_id):
    c = list(db.get_competitor(id=competitor_2_id))
    assert len(c) == 1
    assert c[0].id == competitor_2_id
    assert c[0].first_name == "Angela"
    assert c[0].last_name == "Merkel"
    assert c[0].club_id == club_id
    assert c[0].gender == "F"
    assert c[0].year == 1957
    assert c[0].chip == "1234567"


def test_update_first_added_competitor(db, competitor_1_id, competitor_2_id, club_id):
    db.update_competitor(
        id=competitor_1_id,
        first_name="Anton",
        last_name="Berkel",
        club_id=club_id,
        gender="M",
        year=1958,
        chip="",
    )
    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == competitor_1_id
    assert c[0].first_name == "Anton"
    assert c[0].last_name == "Berkel"
    assert c[0].club_id == club_id
    assert c[0].club_name == "OL Bundestag"
    assert c[0].gender == "M"
    assert c[0].year == 1958
    assert c[0].chip == ""
    assert c[1].id == competitor_2_id
    assert c[1].first_name == "Angela"
    assert c[1].last_name == "Merkel"
    assert c[1].club_id == club_id
    assert c[1].club_name == "OL Bundestag"
    assert c[1].gender == "F"
    assert c[1].year == 1957
    assert c[1].chip == "1234567"


def test_update_last_added_competitor(db, competitor_1_id, competitor_2_id, club_id):
    db.update_competitor(
        id=competitor_2_id,
        first_name="Anton",
        last_name="Berkel",
        club_id=None,
        gender="M",
        year=1958,
        chip="",
    )
    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == competitor_2_id
    assert c[0].first_name == "Anton"
    assert c[0].last_name == "Berkel"
    assert c[0].club_id == None
    assert c[0].club_name == None
    assert c[0].gender == "M"
    assert c[0].year == 1958
    assert c[0].chip == ""
    assert c[1].id == competitor_1_id
    assert c[1].first_name == "Jogi"
    assert c[1].last_name == "Löw"
    assert c[1].club_id == None
    assert c[1].club_name == None
    assert c[1].gender == "M"
    assert c[1].year is None
    assert c[1].chip == ""


def test_add_competitor_with_same_first_name(db, competitor_1_id, club_id):
    competitor_2_id = db.add_competitor(
        first_name="Jogi",
        last_name="Berkel",
        club_id=None,
        gender="M",
        year=1958,
        chip="",
    )
    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == competitor_2_id
    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Berkel"
    assert c[0].club_id == None
    assert c[0].club_name == None
    assert c[0].gender == "M"
    assert c[0].year == 1958
    assert c[0].chip == ""
    assert c[1].id == competitor_1_id
    assert c[1].first_name == "Jogi"
    assert c[1].last_name == "Löw"
    assert c[1].club_id == None
    assert c[1].club_name == None
    assert c[1].gender == "M"
    assert c[1].year is None
    assert c[1].chip == ""


def test_add_competitor_with_same_last_name(db, competitor_1_id, club_id):
    competitor_2_id = db.add_competitor(
        first_name="Norbert",
        last_name="Löw",
        club_id=None,
        gender="M",
        year=1958,
        chip="",
    )
    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == competitor_1_id
    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id == None
    assert c[0].club_name == None
    assert c[0].gender == "M"
    assert c[0].year is None
    assert c[0].chip == ""
    assert c[1].id == competitor_2_id
    assert c[1].first_name == "Norbert"
    assert c[1].last_name == "Löw"
    assert c[1].club_id == None
    assert c[1].club_name == None
    assert c[1].gender == "M"
    assert c[1].year == 1958
    assert c[1].chip == ""


def test_delete_first_added_competitor(db, competitor_1_id, competitor_2_id, club_id):
    db.delete_competitor(id=competitor_1_id)
    c = list(db.get_competitors())
    assert len(c) == 1
    assert c[0].id == competitor_2_id
    assert c[0].first_name == "Angela"
    assert c[0].last_name == "Merkel"
    assert c[0].club_id == club_id
    assert c[0].club_name == "OL Bundestag"
    assert c[0].gender == "F"
    assert c[0].year == 1957
    assert c[0].chip == "1234567"


def test_delete_last_added_competitor(db, competitor_1_id, competitor_2_id, club_id):
    db.delete_competitor(id=competitor_2_id)
    c = list(db.get_competitors())
    assert len(c) == 1
    assert c[0].id == competitor_1_id
    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id == None
    assert c[0].club_name == None
    assert c[0].gender == "M"
    assert c[0].year is None
    assert c[0].chip == ""


def test_add_existing_name_raises_exception(db, competitor_1_id):
    with pytest.raises(repo.ConstraintError, match="Competitor already exist"):
        db.add_competitor(
            first_name="Jogi",
            last_name="Löw",
            club_id=None,
            gender="M",
            year=1958,
            chip="",
        )


def test_change_to_existing_name_raises_exception(db, competitor_1_id, competitor_2_id):
    with pytest.raises(repo.ConstraintError, match="Competitor already exist"):
        db.update_competitor(
            id=competitor_1_id,
            first_name="Angela",
            last_name="Merkel",
            club_id=None,
            gender="F",
            year=None,
            chip="",
        )


def test_update_with_unknown_id_raises_exception(db, competitor_1_id):
    with pytest.raises(KeyError):
        db.update_competitor(
            id=competitor_1_id + 1,
            first_name="Anton",
            last_name="Berkel",
            club_id=None,
            gender="M",
            year=1958,
            chip="",
        )


def test_delete_competitor_with_unknown_id_do_not_change_anything(db, competitor_1_id):
    db.delete_competitor(id=competitor_1_id + 1)
    c = list(db.get_competitors())
    assert len(c) == 1
    assert c[0].id == competitor_1_id
    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id == None
    assert c[0].club_name == None
    assert c[0].gender == "M"
    assert c[0].year is None
    assert c[0].chip == ""


def test_delete_competitor_used_in_entry_raises_exception(
    db, entry_id, competitor_1_id
):
    with pytest.raises(repo.CompetitorUsedError):
        db.delete_competitor(id=competitor_1_id)
